#ifndef ISOCKET_H
#define ISOCKET_H

#include "FamillesAdresse.h"
#include "FamillesProtocole.h"
#include "Port.h"
#include "Adresse.h"
#include "GestionnaireSockets.h"
#include <vector>
#include <string>
#include <utility>
#include <algorithm>
#include <iterator>

namespace SocketsFlux
{
   //
   // Cas d'exception possibles
   //
   class ErreurCreation {};

   //
   // ISocket est une abstraction cachant les dtails
   // d'implantation des sockets sur une plateforme donne
   //
   class ISocket
   {
   protected:
      enum Etat { NonBloquant, Bloquant };
      void etat(const Etat e) throw()
         { etat_ = e; }
      Etat etat() const throw ()
         { return etat_; }
   private:
      Etat etat_;
   public:
      bool est_bloquant() const throw()
         { return etat() == Bloquant; }
      bool est_non_bloquant() const throw()
         { return etat() == NonBloquant; }
      virtual void rendreBloquant() = 0;
      virtual void rendreNonBloquant() = 0;
      ISocket(const Etat e = NonBloquant) // throw(ErreurCreation)
         : etat_(e)
      {
      }
      typedef int erreur_type;
      // Mcanique de connexion
      enum EtatConnexion
      {
         ConnexionOk, ConnexionErreur, ConnexionEnCours
      };
      virtual std::pair<EtatConnexion, erreur_type> connecter
         (const Adresse&, const Port&) = 0;
      class FermetureHomologue
      {
         bool propre_;
      public:
         FermetureHomologue(const bool propre = true)
            : propre_(propre)
         {
         }
         bool propre() const throw()
            { return propre_; }
         bool brutale() const throw()
            { return !propre(); }
      };
      virtual bool fermer() throw()
         { return false; } // Patch!
      virtual ~ISocket() throw()
         { fermer(); }
      virtual bool lier(const Port&) = 0;
      // Mcanique d'acceptation
      enum EtatAcceptation
      {
         AcceptationOk, AcceptationErreur, AcceptationPersonne
      };
      virtual std::pair<EtatAcceptation, ISocket*> accepter() = 0;
      template <class T, class Itt>
         int envoyer(Itt debut, Itt fin)
         {
            const std::size_t NELEMS_TAMPON_DEFAUT = 1024*8;
            return envoyer<T, Itt, NELEMS_TAMPON_DEFAUT>(debut, fin);
         }
      template <class T, class Itt, size_t NELEMS_MAX>
         int envoyer(Itt debut, Itt fin)
         {
            T tampon[NELEMS_MAX];
            const std::size_t NELEMS =
               std::min(static_cast<size_t>(fin - debut), NELEMS_MAX);
            std::copy(debut, debut + NELEMS, tampon);
            return envoyerPrimitif(reinterpret_cast<const char*>(tampon), static_cast<int>(NELEMS*sizeof(T)));
         }
      virtual int envoyerPrimitif(const char *seq, const int n) = 0;
         // throw(std::pair<FermetureHomologue, erreur_type>)
      template <class T, class Itt>
         int recevoir(Itt debut, Itt fin)
         {
            const std::size_t NELEMS_TAMPON_DEFAUT = 1024*8;
            return recevoir<T, Itt, NELEMS_TAMPON_DEFAUT>(debut, fin);
         }
      template <class T, class Itt, size_t NELEMS_MAX>
         int recevoir(Itt debut, Itt fin)
         {
            T tampon[NELEMS_MAX];
            const std::size_t NELEMS =
               std::min(static_cast<std::size_t>(std::distance(debut, fin)), NELEMS_MAX);
            const int NRECUS = recevoirPrimitif(reinterpret_cast<char*>(tampon), NELEMS*sizeof(T));
            std::copy(tampon, tampon+ NRECUS, debut);
            return NRECUS;
         }
         virtual int recevoirPrimitif(char *, const std::size_t) = 0;
         // throw(std::pair<FermetureHomologue,erreur_type>)
      //
      // fabrique
      //
      static ISocket *Creer(const FamilleAdresse, const FamilleProtocole);
   };
}

#endif
